#!/usr/bin/python3
# System modules
import sys, os
import threading
import time
import datetime
import json
import yaml
import requests
import logging
import signal
from elasticsearch import Elasticsearch
from http.server import HTTPServer, BaseHTTPRequestHandler
from socketserver import ThreadingMixIn
import argparse
# to disable ssl errors
from urllib3.exceptions import InsecureRequestWarning
from requests_futures.sessions import FuturesSession
from pprint import pprint
from random import uniform as rand
from Adafruit_IO import Client

logging.basicConfig(level=logging.DEBUG,
                    format='(%(threadName)-10s) %(message)s',
                    )

CFG = None
DATA = {}
DEBUG = False
TEAM = 0
USE_ADA = False


class Handler(BaseHTTPRequestHandler):
    server_version = 'USYS EMBEDDED'
    sys_version = '2.33.1.5'

    def write_headers(self, additional_headers={}):
        state = "ready"
        self.send_header("Content-type", "application/json")
        self.send_header("Access-Control-Allow-Origin", "*")
        self.send_header("X-ICS-State", state)
        if len(additional_headers) > 0:
            for header, value in additional_headers.items():
                try:
                    self.send_header(str(header), value)
                except Exception:
                    logging.warning("failed to write special header")

    def write_body(self, j=None):
        r = False
        if j is not None:
            if 'token' not in j:
                return b'ready.'
            try:
                return (json.dumps(DATA)).encode()
            except KeyError as e:
                logging.error("unable to write json: %s" % str(e))
                return b'error.'
        return b'ready.'

    def do_GET(self):
        self.send_response(200)
        self.write_headers()
        self.end_headers()
        self.wfile.write(self.write_body())

    def do_POST(self):
        content_len = int(self.headers.get('Content-Length'))
        json_data = None
        post_body = self.rfile.read(content_len)
        try:
            json_data = json.loads(post_body.decode('utf8'))
            if json_data is not None:
              for sensor_element in json_data:
                  DATA[sensor_element] = json_data[sensor_element]
        except Exception as e:
            logging.error("unable to load json (%s)" % str(e))
            self.send_error(500, "JSON Payload Bad")
        self.send_response(200)
        self.write_headers()
        self.end_headers()
        self.wfile.write(self.write_body(json_data))



class ThreadingSimpleServer(ThreadingMixIn, HTTPServer):
    pass


class NodeSimulator(object):

    def __init__(self):
        self.last_values = {}  # we can store previous values here, to average? doubt its useful

    def unknown_sim(self, base, target, step):
        final_list = {}
        final_list['data'] = rand(0, 100)

    def power_sim(self, base, target, step):
        final_list = {}
        for sensor_element in CFG['components']:
            if base in sensor_element:
                final_list[sensor_element] = bool(rand(0, 1))
        return final_list

    def collector_sim(self, base, target, step):
        final_list = {}
        # we can do target+/-(value_low+(value_high%target)*step)
        # if target > 0:
        #	 value_low = rand(0, target)
        #	 value_high = rand(target, target + target)
        step_multiplier = 0.0
        # logging.info("simulator base %s: target %d, step: %s"%(base,target,step))
        for sensor_element in CFG['components']:
            if base in sensor_element:
                final_value = 0.0
                if float(step) > 0.0:
                    step_multiplier = rand(0, step)
                else:
                    step_multiplier = rand(step, 1)
                if rand(0, 1) == 0:
                    # subtract
                    final_value = float(target) - float(step_multiplier)
                else:
                    # add
                    final_value = float(target) + float(step_multiplier)
                # alright lets build it all
                final_list[sensor_element] = final_value
        return final_list


class NodeIO:
    def __init__(self):
        self.last_sync = None
        self.last_values = {}
        if CFG['aio_key'] is not None:
            self.aio = Client(CFG['aio_user'], CFG['aio_key'])
        else:
            self.aio = None

    def update(self):
        if self.last_sync is not None and (time.time() - self.last_sync) < 500.0:
            return self.last_values
        try:
            self.last_sync = time.time()
            groups = self.aio.groups()
            for g in groups:
                print('Group {0} has {1} feed(s).'.format(g.name, len(g.feeds)))
                for feed in g.feeds:
                    name = "%s_%s" % (g.name, feed.name)
                    dv = self.aio.data(feed.key)[0]
                    v = None
                    if str(dv.value) == "ON":
                        v = True
                    elif str(dv.value) == "OFF":
                        v = False
                    else:
                        v = float(dv.value)
                    logging.info("Storing %s with value: %s" % (name, v))
                    self.last_values[name] = v
        except Exception as e:
            logging.error("unable top update data %s" % str(e))
        return self.last_values

    def query(self, key):
        query_dict = {}
        self.update()
        pprint(self.last_values)
        try:
            for k, v in self.last_values.items():
                if key in k:
                    query_dict[k] = v
            return query_dict
        except KeyError as e:
            logging.error("unable to query data %s" % str(e))

    def sync(self):
        for k, v in self.last_values:
            DATA[k] = v


class NodePoller(object):
    def __init__(self):
        self.session = None
        self.USE_FUTURE_SESSION = False
        try:
            self.session = FuturesSession()  # switch this to Session()
            self.USE_FUTURE_SESSION = True
        except Exception as e:
            logging.error("An error was encountered when trying to load ")
            self.session = Session()
        self.session.verify = False
        requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

    def query_node(self, base, url):
        request = self.session.get(url)
        result = None
        try:
            rf = self.session.get(url)
            r = rf.result()
            if r.status_code == 200:
                result = r.json()
            else:
                logging.error(
                    "unable process node request due to non-200 http status code: got %d" % int(r.status_code))
        except Exception as e:
            if DEBUG:
                logging.error("unable to process due to exception: %s" % str(e))
        result_list = {}
        if result is not None:
            for sensor_element in result:
                full_sensor_name = sensor_element
                result_list[full_sensor_name] = result[sensor_element]
        return result_list

    def query_ise(self, base, url):
        # i presume we're going to take raw values from here and put them into the values storage
        # park_ and other components
        requests.packages.urllib3.disable_warnings()
        result = None
        if 'team' not in CFG:
            return None
        TEAM = CFG['team']
        try:
            r = self.session.post(url=url, json={'token': '8204h9f385hg0302f802ef20', 'team': TEAM})
            if r.status_code == 200:
                result = r.json()
            else:
                logging.error(
                    "unable process node request due to non-200 http status code: got %d" % int(r.status_code))
        except Exception as e:
            logging.error("Unable to request ISE Results: %s" % str(e))
        return result


class DataCollector():
    def ___init__(self):
        self.run_prod = False
        self.session = FuturesSession()
        self.run = True
        self.debug = False

    def interrupt_handler(self, sig, frames):
        if sig == signal.SIGHUP:
            logging.info("reloading components")
            self.reset()
        else:
            print("Interrupt Reached... Attempting to Shutdown")
            RUN = False
            self.run = RUN
            try:
                requests.get("%s%s" % ("http://", CFG['webserver'].replace("0.0.0.0", "127.0.0.1")))
            except KeyError as e:
                logging.error("shutdown request failed")

    def reset(self):
        with open(config_file, 'r') as stream:
            cfg_data = yaml.load(stream)
            CFG = cfg_data

    def sendToElastic(self, url, payload):
        try:
            pl = {}
            for k,v in payload.items():
                if type(v) == bool or type(v) == int or type(v) == float:
                    pl[k]=int(v)
            es = Elasticsearch(url,verify_certs=False,use_ssl=True)
            pl["@timestamp"] = datetime.datetime.utcnow().isoformat()
            result = es.index(index='park_data', doc_type="doc", body=pl)
            return result
        except KeyError as e:
            logging.error("unable to send data to %s!" % url)
            return 0

    def sendToCentralElastic(self, url, p, add_prefix=False):
        if 'team' not in CFG:
            return None

        try:
            es = Elasticsearch(
                url,
                use_ssl=True,
                verify_certs=False,
                http_auth=('datadumper', 'SP42m21N1O312nsoi32')
            )
        except Exception as e:
            logging.error("unable to login to %s!" % url)
            return 0

        payload = {}
        payload['team'] = CFG['team']
        for key, value in p.items():
            new_key = key
            if add_prefix:
                new_key = "%s_%s" % (CFG['team'], key)
            payload[new_key] = int(value)

        payload["@timestamp"] = datetime.datetime.utcnow().isoformat()

        print("---------------------------------------------------------------")
        print(payload)
        print("---------------------------------------------------------------")

        try:
            result = es.index(index='park_data', doc_type="doc", body=payload)
            return result
        except Exception as e:
            logging.error("unable to send data to %s!" % url)
            return 0

    def runWebServer(self):
        if 'webserver' not in CFG:
            return None
        listener = CFG['webserver'].split(":")
        server = ThreadingSimpleServer((listener[0], int(listener[1])), Handler)
        if bool(CFG['webserver_ssl']):
            import ssl
            server.socket = ssl.wrap_socket(server.socket, keyfile=CFG['webserver_ssl_key'],
                                            certfile=CFG['webserver_ssl_crt'], server_side=True)
        t = threading.Thread(target=server.serve_forever)
        t.daemon = True
        t.start()
        return server

    def run(self):
        # we need to determine if we're running simulators or polling
        active_nodes = {}  # store node_name: node_function
        USE_FUTURE_SESSION = False
        session = None
        # startup our core classes
        np = NodePoller()
        ns = NodeSimulator()
        ni = NodeIO()
        for node_name, node_settings in CFG['nodes'].items():
            if "simulator" in node_settings['mode']:
                if "type" in node_settings:
                    func = "%s_sim" % str(node_settings['type'])
                    active_nodes[node_name] = getattr(ns, func, 'unknown_sim')
            elif "io" in node_settings['mode']:
                active_nodes[node_name] = ni.query
            elif "polling" in node_settings['mode']:
                # we do future requests since its async
                if node_settings['type'] == 'ise':
                    active_nodes[node_name] = np.query_ise
                else:
                    active_nodes[node_name] = np.query_node

        # start webserver for allowing us to see data and read it into other applications
        server = self.runWebServer()
        signal.signal(signal.SIGINT, self.interrupt_handler)

        while self.run:
            if CFG is None:
                logging.error("config is still empty and we cannot continue")
                sys.exit(1)
            result = {}
            # now we start our loop and do things
            for node, node_func in active_nodes.items():
                node_mode = CFG['nodes'][node]['mode']
                if "simulator" in node_mode:
                    if 'target' in CFG['nodes'][node]:
                        t = CFG['nodes'][node]['target']
                    else:
                        t = 50  # can be anything i guess
                    if 'step' in CFG['nodes'][node]:
                        s = CFG['nodes'][node]['step']
                    else:
                        s = 0.5  # could be anything really
                    r = node_func(node, t, s)
                else:
                    u = CFG['nodes'][node]['url']
                    r = node_func(node, u)
                # now that we have data
                pprint(r)
                for element, value in r.items():
                    if DEBUG:
                        logging.info("%s - adding: %s => %s" % (str(node), str(element), str(value)))
                    result[element] = value
                    DATA[element] = value
                time.sleep(0.5)
            if 'team' in CFG:
                DATA['team'] = CFG['team']
            # once copied, send to each elastic
            # in blocking fashion
            if DEBUG:
                logging.info("attempting to send data to central elastic instance")
                #self.sendToCentralElastic(CFG['master_endpoint'], DATA)
            for esc in CFG['endpoints']:
                self.sendToElastic(esc, DATA)
            # wait?
            if 'wait' in CFG:
                if DEBUG:
                    logging.debug("reached wait point of %d seconds" % CFG['wait'])
                time.sleep(CFG['wait'])
            else:
                time.sleep(1)

        try:
            server.shutdown()
        except Exception as e:
            logging.error("warning background web server may not have exited properly: %s" % str(e))
        exit_wait = 2
        if DEBUG:
            print(DEBUG)
            logging.info("waiting %d seconds to exit to ensure all processes have finished" % exit_wait)
        time.sleep(exit_wait)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--config', help='set the configuration definitions file to this file', required=True)
    parser.add_argument('-d', '--debug', action='store_true', help='turn on debugging')
    parser.add_argument('-n', '--nofork', action='store_true', help='do not run as a daemon')
    args = parser.parse_args()
    config_file = str(args.config)

    if not os.path.exists(config_file):
        logging.error("please provide a valid configuration file")
        sys.exit(1)

    if args.debug:
        logging.info("debug has been enabled")
        DEBUG = True

    # yaml examples recommend using with...
    with open(config_file, 'r') as stream:
        cfg_data = yaml.load(stream)
    CFG = cfg_data

    collector = DataCollector()

    if args.nofork is None:
        pid = cfg_data['pidfile']
        if DEBUG:
            logging.info("data collector started")
        daemon = Daemonize(app='datacollector', pid=pid, action=collector.run())
        daemon.start()
    else:
        if DEBUG:
            logging.info("datalogger is starting from config '%s' as foreground instance" % (config_file))
        collector.run()



